/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993,1994 by Mark Wahl
   For license details see documentation
   Procedures for serial port (serial.c)
   created: Mark Wahl DL4YBG 93/07/24
   updated: Mark Wahl DL4YBG 94/12/21
*/

#include "tnt.h"

/* number of frames fetched from monitor before changing to other channel */
#define NUM_MONFRAMES 30
/* number of poll-cycles before data is resent to a former busy TNC */
#define WAIT_BUSY 40

/* actual channel and mode */
extern int act_channel;
extern int act_mode;
extern char tnt_lockfile[];
extern int cookie_flag;
extern int ctext_flag;
extern char resy_log_file[];

/* status of all channels */
struct channel_stat ch_stat[CHANNELS];
/* file descriptor for serial interface */
int serial;
/* file descriptor for lockfile */
int lock;
/* pause-flags for all channels */
int pause_flag[CHANNELS];
/* flag if resync active */
int resync;
/* flag if tx-queue full */
int tx_hiwater;
/* flag if tnc busy */
int tnc_busy;
/* counter for resyncs */
int resync_count;
/* SSID of mailbox */
int tnt_box_ssid;

#ifndef HAS_USLEEP
  struct timeval timevalue;
#endif

/* txqueue for blocks to tnc */
static struct block_buffer block_buffer[TXQUEUE];
static int block_buffer_len;	/* number of blocks in queue */
static int block_buffer_pos;   /* position of first waiting block */

static int rcv_stat;
static int rcv_len;
static int rcv_pos; 
static char txbuffer[259];
static char txbuffer_save[259];
static char rxbuffer[259];
static int poll_channel;
static int req_info;
static int req_flag;
static int req_info_save;
static int req_flag_save;
static int busy_count;
static int len_save;
static int ok_display;
static struct termios org_termios;
static struct termios wrk_termios;

static char hostmode_str[] = "\021\030\022\030\033JHOST 1\015"; 
                             /* ^Q^X^R^XESC at the beginning */
static char hostmode_exit[] = "\0\001\006JHOST 0";

static char resync_string[] = "\001\002INVALID COMMAND: ";

static void send_block();


/* wait time us (linux: time raster 10 ms) */
void uwait(unsigned long time)
{
#ifdef HAS_USLEEP
  usleep(time); /* sleep for 10ms */
#else
  timevalue.tv_usec = time; /* sleep for 10ms */
  timevalue.tv_sec = 0;
  select(0,(struct fd_set *) 0,
           (struct fd_set *) 0,
           (struct fd_set *) 0,
           &timevalue);
#endif
}

/* find a not connected channel, return -1 if none found */
int find_free_channel()
{
  int i;
  
  for (i=1;i<CHANNELS;i++) {
    if ((ch_stat[i].state == 0) && (!xconnect_active(i))) {
      return(i);
    }
  }
  return(-1);
}

/* init of tx-queue */
static void restart_tx_queue()
{
  int i;
  
  tx_hiwater = 0;
  block_buffer_len = 0;
  block_buffer_pos = 0;
  for (i = 0; i < CHANNELS; i++) {
    ch_stat[i].snd_queue_frms = 0;
  }
}

/* opens serial port and set tnc to hostmode
   serstr holds serial device name,
   speed holds baudrate */
int init_serial(serstr,speed,speedflag,unlock)
char *serstr;
unsigned int speed;
int speedflag;
int unlock;
{
  int l;
  char c;
  int len;
#ifdef USE_HIBAUD
  struct serial_struct ser_io;
#endif
  
  if (unlock) unlink(tnt_lockfile);
  if ((lock = open(tnt_lockfile,O_CREAT|O_EXCL,0600)) == -1) {
    printf("Error: serial port locked by other user\n");
    return(1);
  }
  if ((serial = open(serstr,O_RDWR)) == -1) {
    printf("Error: can't open serial port\n");
    close(lock);
    unlink(tnt_lockfile);
    return(1);
  }
  tcgetattr(serial,&org_termios);
#ifdef USE_HIBAUD
  if (speed == B38400) {
    if (ioctl(serial,TIOCGSERIAL, &ser_io) < 0) {
      printf("Error: can't get serial info\n");
      close(serial);
      close(lock);
      unlink(tnt_lockfile);
      return(1);
    }
  }
#endif
  wrk_termios = org_termios;
  wrk_termios.c_cc[VTIME] = 0;
  wrk_termios.c_cc[VMIN] = 0;
  wrk_termios.c_iflag = 0;
  wrk_termios.c_iflag |= IGNBRK;
  wrk_termios.c_oflag = 0;
  wrk_termios.c_lflag = 0;
  wrk_termios.c_cflag |= (CS8|CREAD|CLOCAL);
#ifdef HAS_CRTSCTS
  wrk_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|CRTSCTS|HUPCL);
#else
  wrk_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|HUPCL);
#endif
  cfsetispeed(&wrk_termios,speed);
  cfsetospeed(&wrk_termios,speed);
#ifdef USE_HIBAUD
  if (speed == B38400) {
    ser_io.flags &= ~ASYNC_SPD_MASK;
    ser_io.flags |= speedflag;
    if (ioctl(serial,TIOCSSERIAL, &ser_io) < 0) {
      printf("Error: can't set serial info\n");
      tcsetattr(serial,TCSADRAIN,&org_termios);
      close(serial);
      close(lock);
      unlink(tnt_lockfile);
      return(1);
    }
  }
#endif
  tcsetattr(serial,TCSADRAIN,&wrk_termios);
  
  /* put tnc to hostmode */
  len = sizeof(hostmode_str) - 1;
  if (write(serial,hostmode_str,len) != len){
    printf("Error: can't write to serial port\n");
    return(1);
  }
  sleep(1);
  do {
    l = read(serial,&c,1);
  } while (l);
  printf("TNC in Hostmode\n");
#ifdef USE_SELECT
  wrk_termios.c_cc[VTIME] = 0;
  wrk_termios.c_cc[VMIN] = 1;
  tcsetattr(serial,TCSADRAIN,&wrk_termios);
#endif
  /* clear pause-flag, cookie-send and mycall for all channels */
  /* clear display and log call and set start time */
  for (l = 0; l < CHANNELS; l++) {
    pause_flag[l] = 0;
    ch_stat[l].sendcook = 0;
    ch_stat[l].contcon = 0;
    ch_stat[l].mycall[0] = '\0';
    ch_stat[l].log_call[0] = '\0';
    ch_stat[l].disp_call[0] = '\0';
    ch_stat[l].start_time = time(NULL);
    ch_stat[l].state = 0;
    ch_stat[l].oldstate = 0;
    ch_stat[l].maxstate = 0;
    ch_stat[l].ignore_state = 0;
  }
  resync_count = 0; /* no resyncs */
  tnc_busy = 0; /* TNC not busy */
  return(0);
}

/* put tnc to terminalmode and close serial port */
int exit_serial()
{
  int l;
  char c;
  int len;

#ifdef USE_SELECT
  wrk_termios.c_cc[VTIME] = 0;
  wrk_termios.c_cc[VMIN] = 0;
  tcsetattr(serial,TCSADRAIN,&wrk_termios);
#endif
  /* get tnc back to terminalmode */
  len = sizeof(hostmode_exit);
  if (write(serial,hostmode_exit,len - 1) != len - 1){
    printf("Error: can't write to serial port\n");
    return(1);
  }
  sleep(1);
  do {
      l = read(serial,&c,1);
  } while (l);
  printf("TNC in terminalmode\n");
  tcsetattr(serial,TCSADRAIN,&org_termios);
  close(lock);
  unlink(tnt_lockfile);
  return(0);
}

/* try to resync tnc
   first up to 256 CNTL_A, until tnc responds
   second put tnc to hostmode and then up to 256 CNTL_A, until tnc responds
*/
void resync_tnc(state)
int *state;
{
  int l;
  char c;
  int len;
  int answer;
  int cnt;
  FILE *fp;
  struct tm cvtime;
  time_t ctime;
  char rec_buffer[256];
  int success;

  resync_count++;
  fp = NULL;
  if (resy_log_file[0] != '\0') {  
    fp = fopen(resy_log_file,"a");
  }
  if (fp != NULL) {
    ctime = time(NULL);
    cvtime = *localtime(&ctime);
    fprintf(fp,"%2.2u.%2.2u.%2.2u %2.2u:%2.2u: Resync started\n",
          cvtime.tm_mday,cvtime.tm_mon+1,cvtime.tm_year,
          cvtime.tm_hour,cvtime.tm_min);
  }
  resync = 1;
  success = 0;
  statlin_update();
  sleep(1);
#ifdef USE_SELECT
  wrk_termios.c_cc[VTIME] = 0;
  wrk_termios.c_cc[VMIN] = 0;
  tcsetattr(serial,TCSADRAIN,&wrk_termios);
#endif
  cnt = 0;
  *rec_buffer = '\0';
  do { /* send CNTL_A until tnc responds */
    write(serial,"\001",1);
    cnt++;
    do {
      uwait(10000); /* sleep for 10 ms */
      l = read(serial,&c,1);
      if (l) {
        if (fp != NULL)
          fprintf(fp,"<%x>",c);
        strncat(rec_buffer,&c,1);
        if (strncmp(rec_buffer,resync_string,strlen(rec_buffer)) != 0) {
          *rec_buffer = '\0';
        }
      }
    } while ((l) && (strlen(rec_buffer) < strlen(resync_string)));
    if (strncmp(rec_buffer,resync_string,strlen(resync_string)) == 0)
      success = 1;
  } while (!success && (cnt <= 256));
  if (!success) {
    do {
      /* put tnc to hostmode */
      if (fp != NULL)
        fprintf(fp,"\nSwitching TNC to hostmode\n");
      len = sizeof(hostmode_str) - 1;
      write(serial,hostmode_str,len);
      sleep(1);
      do {
        uwait(10000); /* sleep for 10 ms */
        l = read(serial,&c,1);
        if (l) {
          if (fp != NULL)
            fprintf(fp,"<%x>",c);
        }
      } while (l);
      if (fp != NULL)
        fprintf(fp,"\nTry next Resync\n");
      cnt = 0;
      *rec_buffer = '\0';
      do { /* send CNTL_A until tnc responds */
        write(serial,"\001",1);
        cnt++;
        do {
          uwait(10000); /* sleep for 10ms */
          l = read(serial,&c,1);
          if (l) {
            if (fp != NULL)
              fprintf(fp,"<%x>",c);
            strncat(rec_buffer,&c,1);
            if (strncmp(rec_buffer,resync_string,strlen(rec_buffer)) != 0) {
              *rec_buffer = '\0';
            }
          }
        } while ((l) && (strlen(rec_buffer) < strlen(resync_string)));
        if (strncmp(rec_buffer,resync_string,strlen(resync_string)) == 0)
          success = 1;
      } while (!success && (cnt <= 256));
    } while (!success);
  }
  if (fp != NULL)
    fprintf(fp,"\nReading remaining data\n");
  /* read all characters sent by tnc */
  do {
    l = read(serial,&c,1);
    if (l) {
      if (fp != NULL)
        fprintf(fp,"<%x>",c);
    }
  } while (l);
  if (fp != NULL) {
    fprintf(fp,"\nResync successful\n");
    fclose(fp);
  }
#ifdef USE_SELECT
  wrk_termios.c_cc[VTIME] = 0;
  wrk_termios.c_cc[VMIN] = 1;
  tcsetattr(serial,TCSADRAIN,&wrk_termios);
#endif
  rcv_stat = R_CHANNEL;
  resync = 0;
  statlin_update();
  req_info = RQ_CHSTAT;
  poll_channel = 0;
  ok_display = 0;
  send_block(poll_channel,X_COMM,1,"L");
  *state = S_STAT;
}

static void send_block(channel,code,len,data)
int channel;
char code;
int len;
char data[];
{
  int i;
  
  txbuffer[0] = (char)channel;
  txbuffer[1] = code;
  txbuffer[2] = (char)(len - 1);
  memcpy(txbuffer+3,data,len);
  write(serial,txbuffer,len + 3);
}

static int block_receiver(rxbuffer,len,buffer)
char *rxbuffer;
int len;
char *buffer;
{
  int block_ok;
  int i;
  int error;
  FILE *fp;
  struct tm cvtime;
  time_t ctime;
  
  block_ok = 0;  
  i = 0;
  error = 0;
  while ((!block_ok) && (i < len) && (!error)) {
    switch (rcv_stat) {
    case R_CHANNEL:
      if (buffer[i] >= CHANNELS) {
        error = 1;
        rcv_stat = R_CHANNEL;
      }
      else {
        rxbuffer[0] = buffer[i];
        rcv_stat = R_CODE;
        i++;
      }
      break;
    case R_CODE:
      if (buffer[i] > R_CHANDATA) {
        error = 2;
        rcv_stat = R_CHANNEL;
      }
      else {
        rxbuffer[1] = buffer[i];
        i++;
        switch (rxbuffer[1]) {
          case R_SUCC:
            rcv_stat = R_CHANNEL;
            i++;
            block_ok = 1;
            break;
          case R_SUCCMESS:
          case R_FAILMESS:
          case R_LINKSTAT:
          case R_MONI:
          case R_MONIMESS:
            rcv_stat = R_TOZERO;
            rcv_pos = 0;
            break;
          case R_MONIDATA:
          case R_CHANDATA:
            rcv_stat = R_LENGTH;
            break;
        }
      }
      break;
    case R_LENGTH:
      rxbuffer[2] = buffer[i];
      rcv_len = (int)rxbuffer[2] + 1;
      rcv_stat = R_RDDATA;
      rcv_pos = 0;
      i++;
      break;
    case R_TOZERO:
      if ((rxbuffer[2+rcv_pos] = buffer[i]) == '\0') {
        rcv_stat = R_CHANNEL;
        i++;
        block_ok = 1;
      }
      else {
      rcv_pos++;
      i++;
      }
      break;
    case R_RDDATA:
      rxbuffer[3+rcv_pos] = buffer[i];
      rcv_len--;
      rcv_pos++;
      i++;
      if (rcv_len == 0) {
        block_ok = 1;
        i++;
        rcv_stat = R_CHANNEL;
      }
      break;  
    }
  }
  if ((i < len) && (block_ok)) error = 3;
  if (error) {
    fp = NULL;
    if (resy_log_file[0] != '\0') {  
      fp = fopen(resy_log_file,"a");
    }
    if (fp != NULL) {
      ctime = time(NULL);
      cvtime = *localtime(&ctime);
      fprintf(fp,"%2.2u.%2.2u.%2.2u %2.2u:%2.2u / ",
              cvtime.tm_mday,cvtime.tm_mon+1,cvtime.tm_year,
              cvtime.tm_hour,cvtime.tm_min);
      switch (error ) {
      case 1:
        fprintf(fp,"Illegal channel: %x\n",buffer[i]);
        break;
      case 2:
        fprintf(fp,"Illegal command code: %x\n",buffer[i]);
        break;
      case 3:
        fprintf(fp,"Additional data after block: ");
          while (i < len) {
            fprintf(fp,"%x,",buffer[i]);
            i++;
          }
          fprintf(fp,"\n");
        break;
      }
      fclose(fp);
    }
  }
  return(block_ok);
}

static char *next_value(buffer)
char *buffer;
{
  char c;

  do {
    c = *buffer;
    buffer++;
  } while ((c != ' ') && (c != '\0'));
  return(buffer);
}

static int fill_status(channel,buffer)
int channel;
char *buffer;
{
  int temp;
  int changed;
  int oldstate;
  
  changed = 0;
  ch_stat[channel].st_mess = atoi(buffer);
  buffer = next_value(buffer);
  /* only up to 10 monitorframes at one time */
  if ((channel == 0) && ((temp = atoi(buffer)) > 10))
    ch_stat[channel].rcv_frms = NUM_MONFRAMES;
  else
    ch_stat[channel].rcv_frms = atoi(buffer);
  if (channel != 0) {
    buffer = next_value(buffer);
    temp = atoi(buffer);
    if (ch_stat[channel].snd_frms != temp) {
      ch_stat[channel].snd_frms = temp;
      changed = 1;
    }
    buffer = next_value(buffer);
    temp = atoi(buffer);
    if (ch_stat[channel].unacked != temp) {
      ch_stat[channel].unacked = temp;
      changed = 1;
    }
    buffer = next_value(buffer);
    temp = atoi(buffer);
    if (ch_stat[channel].tries != temp) {
      ch_stat[channel].tries = temp;
      changed = 1;
    }
    buffer = next_value(buffer);
    temp = atoi(buffer);
    if (!ch_stat[channel].ignore_state) {
      ch_stat[channel].oldstate = ch_stat[channel].state;
      if (ch_stat[channel].state != temp) {
        ch_stat[channel].oldstate = ch_stat[channel].state;
        ch_stat[channel].state = temp;
        changed = 1;
      }
    }
    else {
      ch_stat[channel].ignore_state = 0;
    }
    if (ch_stat[channel].state > ch_stat[channel].maxstate)
      ch_stat[channel].maxstate = ch_stat[channel].state;
  }
  return(changed);
}

static int block_decoder(rxbuffer,state)
char *rxbuffer;
int *state;
{
  int channel;
  int block_send;
  int res;
  char tmpstr[10];
#ifdef TNT_CHAMBER
  int ssid,i;
#endif
#ifdef USE_IFACE
  int ssid,i;
#endif
   
  block_send = 0; 
  channel = rxbuffer[0];
  switch (rxbuffer[1]) {
  case R_SUCC:
    if (ok_display) {
      cmd_display(req_flag,channel,"Ok",1);
    }
    if (req_info == RQ_OWNCALL) {
      send_block(channel,X_COMM,1,"I");
      req_info = RQ_CURCALL;
      block_send = 1;
    }
    next_command(state);
    break;
  case R_SUCCMESS:
    switch (req_info) {
    case RQ_NORMAL:
      cmd_display(req_flag,channel,rxbuffer+2,1);
      next_command(state);
      break;
    case RQ_CHSTAT:
      res = fill_status(channel,rxbuffer+2);
      if (((res) && (*state != S_STAT)) ||
          (*state == S_STAT)) {
        stat_display(channel);
#ifdef USE_IFACE
        send_cstatus(channel);
#endif 
      }
      /* send cookie if remote station connects */
      if ((ch_stat[channel].oldstate == 0) &&
          (ch_stat[channel].state >= 4)) {
        if (*state != S_STAT) {
          if (cookie_flag || ctext_flag)
            ch_stat[channel].sendcook = 1;
        }
      }
      /* actions if status changes from not connected to connected */
      if ((ch_stat[channel].oldstate < 4) &&
          (ch_stat[channel].state >= 4)) {
        ch_stat[channel].start_time = time(NULL);
        ch_stat[channel].contcon = 1;
        ch_stat[channel].lastcr = 1;
#ifdef USE_IFACE
        conn_boxcut(channel);
#endif  
      }
      /* actions if status changes from connected to not connected */
      if ((ch_stat[channel].oldstate >= 1) &&
          (ch_stat[channel].state == 0) &&
          (ch_stat[channel].maxstate >= 4)) {
        ch_stat[channel].end_time = time(NULL);
        write_log(channel);
#ifdef USE_SOCKET        
        send_pty_goodbye(channel);
#endif
      }
      /* close files and shells if disconnect and set mycall */
      if ((ch_stat[channel].oldstate > 0) &&
          (ch_stat[channel].state == 0) &&
          (*state != S_STAT)) {
        close_shell(channel,0);
        close_rxfile(channel,0);
        close_txfile(channel,0);
        close_xconnect(channel,0);
        clear_pwmode(channel);
#ifdef USE_IFACE
        close_iface_con(channel,1);
        disc_boxcut(channel);
#endif  
        ch_stat[channel].disp_call[0] = '\0';
        ch_stat[channel].log_call[0] = '\0';
        ch_stat[channel].maxstate = 0;
        if (ch_stat[channel].mycall[0] != '\0') {
          strcpy(tmpstr,"I");
          strcat(tmpstr,ch_stat[channel].mycall);
          send_block(channel,X_COMM,strlen(tmpstr),tmpstr);
          req_info = RQ_NORMAL;
          req_flag = M_CMDSCRIPT;
          block_send = 1;
          break;
        }
      }
      if (*state == S_STAT) {
        send_block(channel,X_COMM,1,"C");
        req_info = RQ_CALL;
        block_send = 1;
      }
      else {
        if (!pause_flag[channel]) {
          if ((ch_stat[channel].rcv_frms) || (ch_stat[channel].st_mess)) {
            send_block(channel,X_COMM,1,"G");
            block_send = 1;
          }
        }
      }
      next_sendframe(channel);
      break;
    case RQ_CALL:
      strcpy(ch_stat[channel].call,rxbuffer+2);
      stat_display(channel);
      set_pwmode(channel);
      send_block(channel,X_COMM,1,"I");
      req_info = RQ_CURCALL;
      block_send = 1;
      break;
    case RQ_CURCALL:
      strcpy(ch_stat[channel].curcall,rxbuffer+2);
      if (ch_stat[channel].sendcook) {
#ifdef TNT_CHAMBER
        ssid=0;
        for(i=0;i<strlen(ch_stat[channel].curcall);i++)
          if(ch_stat[channel].curcall[i]=='-') {
            ssid=atoi(&ch_stat[channel].curcall[i+1]);
            break;
          }
        if(ssid >= 7) {
          rem_ctext(0,0,channel,0,M_CONNECT,NULL);
        }
        else {
          cmd_frontend(0,0,channel,0,0,NULL);
        }
#else
#ifdef USE_IFACE      
        ssid=0;
        for(i=0;i<strlen(ch_stat[channel].curcall);i++)
          if(ch_stat[channel].curcall[i]=='-') {
            ssid=atoi(&ch_stat[channel].curcall[i+1]);
            break;
          }
        if(ssid != tnt_box_ssid) {
          rem_ctext(0,0,channel,0,M_CONNECT,NULL);
        }
        else {
          cmd_box(1,0,channel,0,M_CMDSCRIPT,NULL);
        }
#else
        rem_ctext(0,0,channel,0,M_CONNECT,NULL);
#endif
#endif
        ch_stat[channel].sendcook = 0;
      }
      if (!pause_flag[channel]) {
        if ((ch_stat[channel].rcv_frms) || (ch_stat[channel].st_mess)) {
          send_block(channel,X_COMM,1,"G");
          block_send = 1;
        }
      }
      break;
    }
    break;
  case R_FAILMESS:
    if (strcmp(rxbuffer+2,"TNC BUSY - LINE IGNORED") == 0) {
      tnc_busy = 1;
      busy_count = WAIT_BUSY; /* nbr of times before sending resumed */
      req_info_save = req_info;
      req_flag_save = req_flag;
      len_save = (int)(txbuffer[2] + 1); /* get length of packet */
      memcpy(txbuffer_save,txbuffer,len_save + 3); /* save packet */
      statlin_update();
    }
    else {
      cmd_display(req_flag,channel,rxbuffer+2,1);
      close_xconnect(channel,0);
      next_command(state);
    }
    break;
  case R_LINKSTAT:
    data_display(channel,rxbuffer+2);
    ch_stat[channel].st_mess--;
    if (ch_stat[channel].contcon) {
      next_connect(channel);
      ch_stat[channel].contcon = 0;
    }
    if ((ch_stat[channel].sendcook) && (!ctext_flag)) {
      rem_cookie(1,0,channel,0,0,NULL);
      ch_stat[channel].sendcook = 0;
    }
    send_block(channel,X_COMM,1,"C");
    req_info = RQ_CALL;
    block_send = 1;
    break;
  case R_MONI:
    moni_display(channel,rxbuffer+2);
    if (!pause_flag[channel]) {
      if ((--ch_stat[channel].rcv_frms) || (ch_stat[channel].st_mess)) {
        send_block(channel,X_COMM,1,"G");
        block_send = 1;
      }
    }
    break;
  case R_MONIMESS:
    moni_display(channel,rxbuffer+2);
    send_block(channel,X_COMM,1,"G");
    block_send = 1;
    break;
  case R_MONIDATA:
    moni_display_len(channel,rxbuffer+2);
    if (!pause_flag[channel]) {
      if ((--ch_stat[channel].rcv_frms) || (ch_stat[channel].st_mess)) {
        send_block(channel,X_COMM,1,"G");
        block_send = 1;
      }  
    }
    break;
  case R_CHANDATA:
    data_display_len(channel,rxbuffer+2);
    if (!pause_flag[channel]) {
      if ((--ch_stat[channel].rcv_frms) || (ch_stat[channel].st_mess)) {
        send_block(channel,X_COMM,1,"G");
        block_send = 1;
      }
    }
  }
  return(block_send);
}

/* put block to tx-queue */
int queue_cmd_data(channel,code,len,flag,data)
int channel;
char code;
int len;
int flag;
char data[];
{
  int i;
  int free_block_buffer;
  
  if (block_buffer_len + 1 >= TXQUEUE) return(1);
  if ((free_block_buffer = block_buffer_pos + block_buffer_len) >= TXQUEUE) {
    free_block_buffer -= TXQUEUE;
  }
  if ((code == X_DATA) && (channel != 0)) {
    ch_stat[channel].snd_queue_frms++;
    stat_display(channel);
  }
  block_buffer[free_block_buffer].channel = channel;
  block_buffer[free_block_buffer].code = code;
  block_buffer[free_block_buffer].len = len;
  block_buffer[free_block_buffer].flag = flag;
  memcpy(block_buffer[free_block_buffer].data,data,len);
  block_buffer_len++;
  if (block_buffer_len >= TXQUEUE) {
    tx_hiwater = 1;
  }
  return(0);
}

/* sends block out of txqueue */
static int send_cmd_data()
{
  int channel;
  char code;
  int len;
  int flag;
  
  if (block_buffer_len) {
    channel = block_buffer[block_buffer_pos].channel;
    code = block_buffer[block_buffer_pos].code;
    len = block_buffer[block_buffer_pos].len;
    flag = block_buffer[block_buffer_pos].flag;
    req_info = RQ_NORMAL;
    /* update ax25-state if connect command */
    if ((code == X_COMM) && (len > 1)) { 
      switch (block_buffer[block_buffer_pos].data[0]) {
      case 'C':
        if (channel) {
          if (ch_stat[channel].state == 0) {
            ch_stat[channel].state = 1;
            ch_stat[channel].oldstate = 1;
            ch_stat[channel].ignore_state = 1;
            stat_display(channel);
          }
        }
        break;
      case 'I':
        req_info = RQ_OWNCALL;
        break;
      }
    }
    send_block(channel,code,len,block_buffer[block_buffer_pos].data);
    if (++block_buffer_pos == TXQUEUE) block_buffer_pos = 0;
    block_buffer_len--;
    if (tx_hiwater) {
      if (block_buffer_len < TXQUEUE_FREE) {
        tx_hiwater = 0;
      }
    }
    if (code == X_COMM) {
      req_flag = flag;
      return(1);
    }
    else { /* X_DATA */
      req_flag = M_COMMAND;
      if (channel != 0) {
        ch_stat[channel].snd_queue_frms--;
        ch_stat[channel].snd_frms++;
        stat_display(channel);
      }
      return(0);
    }
  }
}

static void poll_tnc(state)
int *state;
{
  poll_channel++;
  if (poll_channel >= CHANNELS) {
    poll_channel = 0;
    if (*state == S_STAT) {
      *state = S_TNT_UP;
      open_upfile(state);
    }
  }
  send_block(poll_channel,X_COMM,1,"L");
  req_info = RQ_CHSTAT;
}

int data_allowed(channel)
int channel;
{
  if (channel == 0)
    return(1);
  else
    return((ch_stat[channel].snd_frms+ch_stat[channel].snd_queue_frms) <
            MAX_FRAMES);
}

int senddata_allowed(channel)
int channel;
{
  if (channel == 0)
    return(1);
  else
    return((ch_stat[channel].snd_frms+ch_stat[channel].snd_queue_frms) <
            MAX_FRAMES - 1);
}

void serial_server(state,buffer,len)
int *state;
char *buffer;
int len;
{
  if (*state == S_INIT) {
    restart_tx_queue();
    rcv_stat = R_CHANNEL;
    req_info = RQ_CHSTAT;
    poll_channel = 0;
    ok_display = 0;
    send_block(poll_channel,X_COMM,1,"L");
    *state = S_STAT;
  }
  else {
    if (block_receiver(rxbuffer,len,buffer)) {
      if (!block_decoder(rxbuffer,state)) {
        if ((block_buffer_len) && (!busy_count)) {
          ok_display = send_cmd_data();
        }
        else {
          if (tnc_busy) { /* was TNC busy? */
            busy_count--; /* one wait less */
            if (busy_count == 0) { /* now resend ignored data */
              tnc_busy = 0;
              statlin_update();
              memcpy(txbuffer,txbuffer_save,len_save + 3);
              write(serial,txbuffer_save,len_save + 3);
              ok_display = (txbuffer_save[1] == X_COMM);
              req_info = req_info_save;
              req_flag = req_flag_save;
              return;
            }
          }
          poll_tnc(state);
          ok_display = 0;
        }
      }  
    }   
  }
}
